.setcpu "65816"
.smart

;Included Files

.include "Header.asm"

;Define Stuff

HVBJOY=		$4212
JOY0=		$4218
JOY1=		$421a
JOY2=		$421c
JOY3=		$421e
JOYSER0=	$4016
JOYSER1=	$4017
WRIO=		$4201

;Button Numbers
RButton=	16
LButton= 	32
XButton= 	64
AButton= 	128

RightDPad=	1
LeftDPad=	2
DownDPad=	4
UpDPad=		8
StartButton=	16
SelectButton=	32
YButton=	64
BButton=	128

WordRightDPad=		256
WordLeftDPad=		512
WordDownDPad=		1024
WordUpDPad=		2048
WordStartButton=	4096
WordSelectButton=	8192
WordYButton=		16384
WordBButton=		32768

;Tile Request Table Format
VramAdress=	0
TileAddress=	2
BankNumber=	4	

;=========================================================================================
;=========================================================================================

.struct ObjectSlot
Identity	.byte 3
XPosition	.byte 3
YPosition	.byte 3
VramOffset	.byte 2
MetaspriteTableOffset	.byte 2
FlipBits	.byte 2
.endstruct

;=========================================================================================
.segment "ZEROPAGE"
;=========================================================================================

Joy1Data:		.res 2
Joy2Data:		.res 2
Joy1Press:		.res 2
Joy2Press:		.res 2
SpriteCount:		.res 2
ObjectTableOffset:	.res 2
NewObjectRequest:	.res 3
ColorCounter:		.res 2
LongJumpLocation:	.res 3
HFlipMask:		.res 2
VFlipMask:		.res 2
MetaspriteCount:	.res 2
MetaspriteAttributes:	.res 2
TileRequestCounter16x16:	.res 2
MetaspriteBaseTileNumber:	.res 2
MetaspriteVramOffset:		.res 2

;SNES RAM area, $0200-1fff
;See snes.cfg

;=========================================================================================
.segment "BSS"
;=========================================================================================

SpriteBuf1:		.res $200	;Sprite buffer for OAM DMA, 512 ($200) bytes
SpriteBuf1Size=		*-SpriteBuf1
SpriteBuf2:    		.res $20	;Sprite buffer for OAM DMA, 32  ($20) bytes
SpriteBuf2Size=		*-SpriteBuf2
SpriteBufSize= 		SpriteBuf1Size+SpriteBuf2Size
SpriteBuf3:		.res $200	;Sprite buffer for OAM DMA, 512  ($200) bytes
SpriteBuf3Size=		*-SpriteBuf3
ObjectSlotSize=		.sizeof(ObjectSlot)
ObjectSlotNumber=	16
ObjectTable:		.res ObjectSlotSize*ObjectSlotNumber
ObjectTableSize=	 *-ObjectTable
TileRequestSlotSize=	6
TileRequestSlotNumber=	128
TileRequestTable:	.res TileRequestSlotSize*TileRequestSlotNumber
VramSlotSize=		6
VramSlotNumber=		128
VramTable:		.res VramSlotSize*VramSlotNumber

;=========================================================================================
.segment "BSS7E"
;=========================================================================================

;Assemble-time assertion; SNES DMA transfer can only handle 65536 bytes.
.assert (SpriteBufSize < $10000), error, "SpriteBufs area too big for SNES DMA transfer"

;=========================================================================================
.segment "CODE"
;=========================================================================================

.proc Main

  sei		;Disable interrupts
  clc		;Clear carry, used by XCE
  xce		;Set 65816 native mode
  jmp :+	;Needed to set K (bank of PC) properly if MODE 21 is ever used;
		;see official SNES developers docs, "Programming Cautions"

: cld		;Disable decimal mode
  phk		;Push K (PC bank)
  plb		;Pull B (data bank, i.e. data bank now equals PC bank)

  rep #$30	;A=16, X/Y=16

;Note: this should correlate with ZEROPAGE in lorom256k.cfg
  lda #$0000
  tcd		;Set D = $0000 (direct page)

;Note: this should correlate with the top of BSS in lorom256k.cfg
  ldx #$1fff
  txs		;Set X = $1fff (stack pointer)

;Register initialisation values, per official Nintendo documentation
;Note: if you are looking at this code and thinking "how can I optimise"
;this?", then you are doing nothing useful with your time.  This code is
;called ONCE on reset or power-on, it does not need to be optimised.  I
;(koitsu) find it remarkable how many people have tried to "optimise"
;this routine.  STOP IT.

  sep #$20	;A=8
  lda #$80
  sta $2100
  stz $2101
  stz $2102
  stz $2103
  stz $2104
  stz $2105
  stz $2106
  stz $2107
  stz $2108
  stz $2109
  stz $210A
  stz $210B
  stz $210C
  stz $210D
  stz $210D
  stz $210E
  stz $210E
  stz $210F
  stz $210F
  stz $2110
  stz $2110
  stz $2111
  stz $2111
  stz $2112
  stz $2112
  stz $2113               
  stz $2113               
  stz $2114
  stz $2114
  lda #$80
  sta $2115
  stz $2116
  stz $2117
  stz $211A
  stz $211B
  lda #$01
  sta $211B
  stz $211C
  stz $211C
  stz $211D
  stz $211D
  stz $211E       
  lda #$01
  sta $211E
  stz $211F
  stz $211F
  stz $2120
  stz $2120
  stz $2121
  stz $2123
  stz $2124
  stz $2125
  stz $2126
  stz $2127
  stz $2128
  stz $2129
  stz $212A
  stz $212B
  stz $212C
  stz $212D
  stz $212E
  stz $212F
  stz $4200
  lda #$FF
  sta $4201
  stz $4202
  stz $4203
  stz $4204
  stz $4205
  stz $4206
  stz $4207
  stz $4208
  stz $4209
  stz $420A
  stz $420B
  stz $420C
  stz $420D

  jsr clear_vram	;Zero VRAM. Requires A=8, X/Y=16
  jsr clear_cgram	;Zero palette. Requires A=8, X/Y=16
  
;clear sprite tables

  stz $2102	;sprites initialized to be off the screen, palette 0, character 0
  stz $2103
  ldx #$0080
  lda #$F0

loop08:
  sta $2104	;set X = 240
  sta $2104	;set Y = 240
  sta $2104	;set character = $00
  sta $2104	;set priority=0, no flips
  dex
  bne loop08
  ldx #$0020

loop09:
  stz $2104		;set size bit=0, x MSB = 0
  dex
  bne loop09

;clear SNES RAM

  stz $2181	;set WRAM address to $000000
  stz $2182
  stz $2183

  ldx #$8008
  stx $4300		;Set DMA mode to fixed source, BYTE to $2180
  ldx #.LOWORD(zero_fill_byte)
  stx $4302		;Set source offset
  lda #.BANKBYTE(zero_fill_byte)
  sta $4304		;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 64KBytes (65536 bytes)
  lda #$01
  sta $420B		;Initiate transfer

  lda #$01		;now zero the next 64KB (i.e. 128KB total)
  sta $420B		;Initiate transfer

  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8

  jsr initialize_oam
  jsr initialize_controllers
  jsr setup_video		;setup the video modes and such, then turn on the screen

;Enable NMI (when interrupts get re-enabled) and joypad reading
  sep #$20	;A=8
  lda #$81
  sta $4200

  cli			;Restore interrupts -- we're all ready to go!
  
  rep #$30	;A=16, X/Y=16
  lda #$000C
  sta TileRequestCounter16x16
  lda #$0000
  sta TileRequestTable+VramAdress
  lda #.LOWORD(Test1Tiles)
  sta TileRequestTable+TileAddress
  sep #$20	;A=8
  lda #.BANKBYTE(Test1Tiles)
  sta TileRequestTable+BankNumber+1

  rep #$30	;A=16, X/Y=16
  lda #$0040
  sta TileRequestTable+VramAdress+6
  lda #.LOWORD(Test2Tiles)
  sta TileRequestTable+TileAddress+6
  sep #$20	;A=8
  lda #.BANKBYTE(Test2Tiles)
  sta TileRequestTable+BankNumber+7

  rep #$30	;A=16, X/Y=16
  lda #.LOWORD(dummy)
  sta NewObjectRequest
  sep #$20	;A=8
  lda #.BANKBYTE(dummy)
  sta NewObjectRequest+2

  rep #$30	;A=16, X/Y=16

  lda #TestBlockMetasprite
  sta ObjectTable+ObjectSlot::MetaspriteTableOffset

  lda #$000C
  sta VramTable+4

  jsr object_spawner

;=========================================================================================

infinite_loop:
  wai			;Wait for interrupts to finish (NMI/VBlank)

  rep #$30	;A=16, X/Y=16

  jsr object_identifier

  rep #$30	;A=16, X/Y=16
  inc ColorCounter
  sep #$20	;A=8
  lda #$00
  sta $2121
  lda ColorCounter
  sta $2122
  lda ColorCounter+1
  sta $2122

  jmp infinite_loop	;Do this forever
.endproc

;=========================================================================================
;=========================================================================================

.proc VBlank
  rep #$30	;A=16, X/Y=16
  pha
  phx
  phy
  phd
  phb
  php
  lda #$0000
  tcd

  jsr tile_uploader

;transfer sprite data

  sep #$20	;A=8
  stz $2102	;set OAM address to 0
  stz $2103	;...and no priority rotation

  stz $4300			;CPU -> PPU, auto-increment, write once
  lda #$04
  sta $4301			;DMA will write to register $2104 (OAM data write)
  ldy #.LOWORD(SpriteBuf1)
  sty $4302			;source offset of SpriteBuf1
  lda #.BANKBYTE(SpriteBuf1)
  sta $4304			;bank address of SpriteBuf1
  ldy #SpriteBufSize
  sty $4305			;transfer SpriteBufSize bytes (pre-calculated number)
  lda #$01
  sta $420B		;start DMA transfer

  jsr get_input	;update joypad data

  lda $4210	;clear NMI Flag

  plp
  plb
  pld
  ply
  plx
  pla
  rti
.endproc

;=========================================================================================
;=========================================================================================

.proc clear_vram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  lda #$80
  sta $2115		;Set VRAM port to word access
  ldx #$1809
  stx $4300		;Set DMA mode to fixed source, WORD to $2118/9
  ldx #$0000
  stx $2116			;Set VRAM port address to $0000
  ldx #.LOWORD(zero_fill_byte)
  stx $4302			;Set source address to $xx:0000
  lda #.BANKBYTE(zero_fill_byte)
  sta $4304			;Set source bank
  ldx #$0000
  stx $4305		;Set transfer size to 65536 bytes
  lda #$01
  sta $420B		;Initiate transfer
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc clear_cgram
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8
  stz $2121
  ldx #$0100

clear_palette_loop:
  stz $2122
  stz $2122
  dex
  bne clear_palette_loop
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc initialize_oam
  php	
  rep #$30

  lda #$01
  ldx #0

initialize_oam_loop:
  sta SpriteBuf3+1,x
  inx
  inx
  inx
  inx
  cpx #$0200
  bne initialize_oam_loop
  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc initialize_controllers
  php	

  rep	#$10	;8 bit mem/A, 16 bit X/Y
  sep	#$20

  lda #$C0	;have the automatic read of the SNES read the first pair of JoyPads
  sta WRIO

  ldx #$0000
  stx Joy1Press
  stx Joy2Press

;in ZSNES, the auto-read doesn't work unless NMI is enabled... is this correct?
;
; lda $4212
; lda $4212
;
;w11:
; lda $4212
; and #$01
; beq _w11	;Wait till automatic JoyPort read starts
;
; jsr GetInput	;fill the variables with real Joypad data
;
;w12:
; lda $4212
; and #$01
; beq _w12	;Wait till automatic JoyPort read starts
;
; jsr get_input	;fill the variables with real Joypad data

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc get_input
  php
  sep #$20	;A=8

  lda #$01
: bit $4212
  bne :-		;Wait till automatic JoyPort read is complete

  rep #$30   ;A=16, X/Y=16

;Get Joypads 1,2

  lda JOY0		;get JoyPad1
  tax 			
  eor Joy1Data		;A = A xor JoyState = (changes in joy state)
  stx Joy1Data		;update JoyState
  ora Joy1Press		;A = (joy changes) or (buttons pressed)
  and Joy1Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy1Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

  lda JOY1		;get JoyPad2
  tax 			
  eor Joy2Data		;A = A xor JoyState = (changes in joy state)
  stx Joy2Data		;update JoyState
  ora Joy2Press		;A = (joy changes) or (buttons pressed)
  and Joy2Data		;A = ((joy changes) or (buttons pressed)) and (current joy state)  
  sta Joy2Press		;store A = (buttons pressed since last clearing reg) and (button is still down)

;Make sure Joypads 1,2 are valid

  sep #$30	;A=8, X/Y=8
  lda JOYSER0
  eor #$01
  and #$01		;A = -bit0 of JoySer0
  ora Joy1Data
  sta Joy1Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

  lda JOYSER1
  eor #$01
  and #$01		;A = -bit0 of JoySer1
  ora Joy2Data		
  sta Joy2Data		;joy state = (joy state) or A.... so bit0 of Joy1State = 0 only if it is a valid joypad

;Change all invalid joypads to have a state of no button presses

  rep #$30	;A=16, X/Y=16

  ldx #$0001
  lda #$000f

  bit Joy1Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq :+		;into the joy port, or it is not a joypad
  stx Joy1Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy1Press

: bit Joy2Data		;A = joy state, if any of the bottom 4 bits are on... either nothing is plugged
  beq done		;into the joy port, or it is not a joypad
  stx Joy2Data		;if it is not a valid joypad put $0001 into the 2 joy state variables
  stz Joy2Press

done:
  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc setup_video
  php
  rep #$30	;A=16, X/Y=16
  sep #$20	;A=8

  stz $211b
  lda #$01
  sta $211b
  stz $211c
  stz $211c
  stz $211d
  stz $211d
  stz $211e
  sta $211e
  lda #$80
  sta $211f
  stz $211f
  sta $2120
  stz $2120

  lda #%00000000	;Sprites 8x8 or 16x16, character data at $0000 (word address)
  sta $2101         

  lda #$01		;Set video mode 1, 16x16 tiles (16 color BG1, 16 color BG2, 4 color BG3)
  sta $2105         

  lda #$70		;Set BG1's Tile Map VRAM offset to $7800 (word address)
  sta $2107		;and the Tile Map size to 32 tiles x 32 tiles

  lda #$02		;Set BG1's Character VRAM offset to $4000 (word address)
  sta $210B

  lda #$10		;Turn on Sprites
  sta $212C

  lda #$0F		;Turn on screen, full brightness
  sta $2100

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc process_oam
  php
  rep #$10
  sep #$20

  ldx SpriteCount
  lda #$01

clear_oam_loop:
  cpx #$0200
  beq finished_clearing_oam_loop
  sta SpriteBuf1,x
  stz SpriteBuf1+1,x
  stz SpriteBuf1+2,x
  stz SpriteBuf1+3,x
  sta SpriteBuf3+1,x		;sprite 3, 9th x bit
  inx
  inx
  inx
  inx
  bra clear_oam_loop

finished_clearing_oam_loop:
  ldy #$0000
  sty SpriteCount

build_hioam_loop:
  lda SpriteBuf3+14,y		;sprite 3, size bit
  asl
  ora SpriteBuf3+13,y		;sprite 3, 9th x bit
  asl
  ora SpriteBuf3+10,y		;sprite 2, size bit
  asl
  ora SpriteBuf3+9,y		;sprite 2, 9th x bit
  asl
  ora SpriteBuf3+6,y		;sprite 1, size bit
  asl
  ora SpriteBuf3+5,y		;sprite 1, 9th x bit
  asl
  ora SpriteBuf3+2,y		;sprite 0, size bit
  asl
  ora SpriteBuf3+1,y		;sprite 0, 9th x bit
  sta SpriteBuf1,x
  rep #$20
  tya
  clc
  adc #$0010
  tay
  sep #$20
  inx
  cpx #$0220
  bne build_hioam_loop

  plp
  rts
.endproc

;=========================================================================================
;=========================================================================================

.proc object_spawner
  rep #$30    ; A=16, X/Y=16
  ldx #$0000

object_spawner_loop:
  lda a:ObjectTable+ObjectSlot::Identity,x
  bne continue_object_spawner_loop
  lda a:ObjectTable+ObjectSlot::Identity+1,x
  beq object_spawner_loop_done
continue_object_spawner_loop:
  txa
  clc
  adc #ObjectSlotSize
  tax
  cpx #ObjectTableSize
  bne object_spawner_loop

object_spawner_loop_done:
  lda a:NewObjectRequest
  sta a:ObjectTable+ObjectSlot::Identity,x
  lda a:NewObjectRequest+1
  sta a:ObjectTable+ObjectSlot::Identity+1,x
  rts

.endproc

;=========================================================================================
;=========================================================================================

.proc object_identifier
  rep #$30	; A=16, X/Y=16
  lda #ObjectTable
  sta ObjectTableOffset
  tcd

object_identifier_loop:
  lda ObjectSlot::Identity
  cmp #$0001
  bcs continue_object_identifier_loop
  lda ObjectSlot::Identity+1
  beq next_object
  lda ObjectSlot::Identity
continue_object_identifier_loop:
  sta a:LongJumpLocation
  lda ObjectSlot::Identity+1
  sta a:LongJumpLocation+1
  jml [LongJumpLocation]
next_object:
  rep #$30	; A=16, X/Y=16
  lda a:ObjectTableOffset		;says how many objects have been identified
  clc
  adc #ObjectSlotSize			;add 24 to look at the next object
  sta a:ObjectTableOffset		;store the result for the next time we go through the loop
  tcd					;transfer the accumulator to y for "cpy #8192" and "lda ObjectTable,y"
  cmp #ObjectTableSize+ObjectTable	;sees if all objects have been identified (each object is 48 bytes)
  bne object_identifier_loop		;if so, quit searching
  jsr metasprite_handler
  lda #$0000
  tcd
  jsr process_oam
  rts

.endproc

;=========================================================================================
;=========================================================================================

.proc dummy
  rep #$30	; A=16, X/Y=16

up:
  lda a:Joy1Data
  and #WordUpDPad
  beq down
  dec ObjectSlot::YPosition
  lda ObjectSlot::FlipBits
  and #$7FFF
  sta ObjectSlot::FlipBits

down:
  lda a:Joy1Data
  and #WordDownDPad
  beq left
  inc ObjectSlot::YPosition
  lda #$7FFF
  lda ObjectSlot::FlipBits
  ora #$8000
  sta ObjectSlot::FlipBits

left:
  lda a:Joy1Data
  and #WordLeftDPad
  beq right
  dec ObjectSlot::XPosition
  lda ObjectSlot::FlipBits
  and #$BFFF
  sta ObjectSlot::FlipBits

right:
  lda a:Joy1Data
  and #WordRightDPad
  beq dummy_done
  inc ObjectSlot::XPosition
  lda ObjectSlot::FlipBits
  ora #$4000
  sta ObjectSlot::FlipBits

dummy_done:
  jmp object_identifier::next_object
.endproc

;=========================================================================================
;=========================================================================================

.proc metasprite_handler
  rep #$30	;A=16, X/Y=16
  lda #ObjectTable
  tcd
  ldy #$0000
  bra continue_metasprite_finder

metasprite_finder:
  tdc
  clc
  adc #ObjectSlotSize
  tcd
  cmp #ObjectTable+ObjectTableSize
  bne continue_metasprite_finder
  sty a:SpriteCount
  rts

continue_metasprite_finder:
  ldx ObjectSlot::MetaspriteTableOffset
  beq metasprite_finder
  lda a:$0000,x
  cmp #$0001
  bne multiple_metasprite
  brl single_metasprite

multiple_metasprite:
  sta a:MetaspriteCount
  phx
  lda ObjectSlot::VramOffset
  sta a:MetaspriteVramOffset
  tax
  lda a:VramAdressConverterTable,x
  sta a:MetaspriteBaseTileNumber
  plx

horizontal_flip_check:
  stz a:HFlipMask
  lda ObjectSlot::FlipBits
  bit #$4000
  beq vertical_flip_check
  lda #$FFFF
  sta a:HFlipMask

vertical_flip_check:
  stz a:VFlipMask
  lda ObjectSlot::FlipBits
  bit #$8000
  beq metasprite_loop
  lda #$FFFF
  sta a:VFlipMask

metasprite_loop:
  cpy #$0200			;sees if all 128 sprites are used up
  bne continue_metasprite_loop1
  brl metasprite_finder

continue_metasprite_loop1:
  lda a:MetaspriteCount
  bne continue_metasprite_loop2
  brl metasprite_finder

continue_metasprite_loop2:
  lda a:$0006,x
  and #$0001
  bne sprite_not_flipped_horizontally
  sta a:SpriteBuf3+2,y		;sprite size
  lda ObjectSlot::FlipBits
  and #$4000
  beq sprite_not_flipped_horizontally
  lda a:$0002,x
  clc
  eor a:HFlipMask
  adc ObjectSlot::XPosition
  clc
  adc #$0008
  cmp #256
  bcc sprite_x_not_out_of_bounds
  cmp #65504
  bcs sprite_x_not_out_of_bounds
  brl sprite_out_of_bounds

sprite_not_flipped_horizontally:
  lda a:$0002,x
  clc
  eor a:HFlipMask
  adc ObjectSlot::XPosition
  cmp #256
  bcc sprite_x_not_out_of_bounds
  cmp #65504
  bcs sprite_x_not_out_of_bounds
  brl sprite_out_of_bounds

sprite_x_not_out_of_bounds:
  and #$01FF
  sta a:SpriteBuf3,y		;Store sprite X position SpriteBuf1+y
  and #$00FF
  sta a:SpriteBuf1,y		;Store sprite X position SpriteBuf1+y
  lda a:$0006,x
  and #$0001
  bne sprite_not_flipped_vertically
  sta a:SpriteBuf3+2,y		;sprite size
  lda ObjectSlot::FlipBits
  and #$8000
  beq sprite_not_flipped_vertically
  lda a:$0004,x			;2nd byte = sprite Y position (value 0-255)
  eor a:VFlipMask
  clc
  adc ObjectSlot::YPosition
  clc
  adc #$0008
  cmp #224
  bcc sprite_y_not_out_of_bounds
  cmp #65504
  bcs sprite_y_not_out_of_bounds
  bra sprite_out_of_bounds

sprite_not_flipped_vertically:
  lda a:$0004,x			;2nd byte = sprite Y position (value 0-255)
  eor a:VFlipMask
  clc
  adc ObjectSlot::YPosition
  cmp #224
  bcc sprite_y_not_out_of_bounds
  cmp #65504
  bcs sprite_y_not_out_of_bounds
  bra sprite_out_of_bounds

sprite_y_not_out_of_bounds:
  and #$00FF
  sta a:SpriteBuf1+1,y
  and #$8000
  lda a:$0006,x
  and #$0001
  sta a:SpriteBuf3+2,y		;sprite size
  lda ObjectSlot::FlipBits
  lda a:$0006,x
  and #$0100
  beq sprite_y_not_out_of_bounds_continued
  phx
  ldx a:MetaspriteVramOffset
  lda a:VramTable+4,x
  tax
  sta a:MetaspriteVramOffset
  lda a:VramAdressConverterTable,x
  sta a:MetaspriteBaseTileNumber
  plx

sprite_y_not_out_of_bounds_continued:
  lda a:$0008,x
  clc
  adc a:MetaspriteBaseTileNumber
  eor ObjectSlot::FlipBits
  sta a:SpriteBuf1+2,y		;extra/character
  iny
  iny
  iny
  iny

sprite_out_of_bounds:
  txa
  clc
  adc #$0008
  tax
  dec a:MetaspriteCount		;decrement MetaspriteCount by 1
  brl metasprite_loop		;back to the loop...

;=========================================================================================

single_metasprite:
  phx
  lda ObjectSlot::VramOffset
  tax
  lda a:VramAdressConverterTable,x
  sta a:MetaspriteBaseTileNumber
  plx

  cpy #$0200			;sees if all 128 sprites are used up
  bne single_metasprite_continued
  brl metasprite_finder
  
single_metasprite_continued:
  and #$00FF
  sta a:SpriteBuf1+1,y		;store sprite Y position in SpriteBuf1+1,y

  lda a:$0002,x			;1st byte = sprite X position (value 0-255)
  clc
  adc ObjectSlot::XPosition
  cmp #256
  bcc single_sprite_x_not_out_of_bounds
  cmp #65504
  bcs single_sprite_x_not_out_of_bounds
  bra single_sprite_out_of_bounds

single_sprite_x_not_out_of_bounds:
  and #$01FF
  sta a:SpriteBuf3,y		;store sprite X position SpriteBuf1+y
  and #$00FF
  sta a:SpriteBuf1,y		;store sprite X position SpriteBuf1+y
  lda a:$0004,x			;2st byte = sprite Y position (value 0-255)
  clc
  adc ObjectSlot::YPosition
  cmp #224
  bcc single_sprite_y_not_out_of_bounds
  cmp #65504
  bcs single_sprite_y_not_out_of_bounds
  bra single_sprite_out_of_bounds

single_sprite_y_not_out_of_bounds:
  sta a:SpriteBuf1+1,y
  lda a:$0006,x
  and #$0001
  sta a:SpriteBuf3+2,y		;sprite size
  lda a:$0008,x
  clc
  adc a:MetaspriteBaseTileNumber
  eor ObjectSlot::FlipBits+1
  sta a:SpriteBuf1+2,y		;extra/character
  iny
  iny
  iny
  iny

single_sprite_out_of_bounds:
  txa
  clc
  adc #$0008
  tax
  dec a:MetaspriteCount		;decrement MetaspriteCount by 1
  brl metasprite_finder
.endproc

;=========================================================================================
;=========================================================================================

tile_uploader:
  rep #$30	; A=16, X/Y=16
  ldx #$0000
  lda #$1801			; Set DMA mode (word, normal increment) and destination register (VRAM write register)
  sta $4300
  sta $4310
  lda #$0080
  sta $2115

tile_uploader_16x16:
  lda TileRequestCounter16x16
  beq tile_uploader_done

  lda #$0040
  sta $4305
  sta $4315

;16x16 Top Half
  lda TileRequestTable+VramAdress,x
  sta $2116

  lda TileRequestTable+BankNumber,x
  sta $4303
  sta $4313

  lda TileRequestTable+TileAddress,x
  sta $4302
  clc
  adc #$0040
  sta $4312

  lda #%0000000000000001	; Initiate DMA transfer (channel 0)
  sta $420B


;16x16 Bottom Half
  lda TileRequestTable+VramAdress,x
  clc
  adc #$0100
  sta $2116

  lda #%0000000000000010	; Initiate DMA transfer (channel 1)
  sta $420B

  txa
  clc
  adc #TileRequestSlotSize
  tax
  cpx TileRequestCounter16x16
  bne tile_uploader_16x16

tile_uploader_done:
  rts

;=========================================================================================
.segment "RODATA"
;=========================================================================================

VramAdressConverterTable:
  .word $0000,$0000,$0000,$0002,$0000,$0000,$0004,$0000,$0000,$0006,$0000,$0000
  .word $0008,$0000,$0000,$000A,$0000,$0000,$000C,$0000,$0000,$000E,$0000,$0000
  .word $0020,$0000,$0000,$0022,$0000,$0000,$0024,$0000,$0000,$0026,$0000,$0000
  .word $0028,$0000,$0000,$002A,$0000,$0000,$002C,$0000,$0000,$002E,$0000,$0000
  .word $0040,$0000,$0000,$0042,$0000,$0000,$0044,$0000,$0000,$0046,$0000,$0000
  .word $0048,$0000,$0000,$004A,$0000,$0000,$004C,$0000,$0000,$004E,$0000,$0000
  .word $0060,$0000,$0000,$0062,$0000,$0000,$0064,$0000,$0000,$0066,$0000,$0000
  .word $0068,$0000,$0000,$006A,$0000,$0000,$006C,$0000,$0000,$006E,$0000,$0000
  .word $0080,$0000,$0000,$0082,$0000,$0000,$0084,$0000,$0000,$0086,$0000,$0000
  .word $0088,$0000,$0000,$008A,$0000,$0000,$008C,$0000,$0000,$008E,$0000,$0000
  .word $00A0,$0000,$0000,$00A2,$0000,$0000,$00A4,$0000,$0000,$00A6,$0000,$0000
  .word $00A8,$0000,$0000,$00AA,$0000,$0000,$00AC,$0000,$0000,$00AE,$0000,$0000
  .word $00C0,$0000,$0000,$00C2,$0000,$0000,$00C4,$0000,$0000,$00C6,$0000,$0000
  .word $00C8,$0000,$0000,$00CA,$0000,$0000,$00CC,$0000,$0000,$00CE,$0000,$0000
  .word $00E0,$0000,$0000,$00E2,$0000,$0000,$00E4,$0000,$0000,$00E6,$0000,$0000
  .word $00E8,$0000,$0000,$00EA,$0000,$0000,$00EC,$0000,$0000,$00EE,$0000,$0000
  .word $0100,$0000,$0000,$0102,$0000,$0000,$0104,$0000,$0000,$0106,$0000,$0000
  .word $0108,$0000,$0000,$010A,$0000,$0000,$010C,$0000,$0000,$010E,$0000,$0000
  .word $0120,$0000,$0000,$0122,$0000,$0000,$0124,$0000,$0000,$0126,$0000,$0000
  .word $0128,$0000,$0000,$012A,$0000,$0000,$012C,$0000,$0000,$012E,$0000,$0000
  .word $0140,$0000,$0000,$0142,$0000,$0000,$0144,$0000,$0000,$0146,$0000,$0000
  .word $0148,$0000,$0000,$014A,$0000,$0000,$014C,$0000,$0000,$014E,$0000,$0000
  .word $0160,$0000,$0000,$0162,$0000,$0000,$0164,$0000,$0000,$0166,$0000,$0000
  .word $0168,$0000,$0000,$016A,$0000,$0000,$016C,$0000,$0000,$016E,$0000,$0000
  .word $0180,$0000,$0000,$0182,$0000,$0000,$0184,$0000,$0000,$0186,$0000,$0000
  .word $0188,$0000,$0000,$018A,$0000,$0000,$018C,$0000,$0000,$018E,$0000,$0000
  .word $01A0,$0000,$0000,$01A2,$0000,$0000,$01A4,$0000,$0000,$01A6,$0000,$0000
  .word $01A8,$0000,$0000,$01AA,$0000,$0000,$01AC,$0000,$0000,$01AE,$0000,$0000
  .word $01C0,$0000,$0000,$01C2,$0000,$0000,$01C4,$0000,$0000,$01C6,$0000,$0000
  .word $01C8,$0000,$0000,$01CA,$0000,$0000,$01CC,$0000,$0000,$01CE,$0000,$0000
  .word $01E0,$0000,$0000,$01E2,$0000,$0000,$01E4,$0000,$0000,$01E6,$0000,$0000
  .word $01E8,$0000,$0000,$01EA,$0000,$0000,$01EC,$0000,$0000,$01EE,$0000,$0000

Test1Tiles:
  .incbin "Test1.pic"

Test2Tiles:
  .incbin "Test2.pic"

TestBlockMetasprite:
  .word $0003	; Number of metasprite table entries below
        ;XPos YPos NextTile/Size Extra/Character
  .word $0000,$0000,$0000,$0001
  .word $0008,$0000,$0001,$0000
  .word $0000,$0010,$0101,$4000

;Byte used to zero/initialise all of memory.
;This is needed because DMA requires an address of a byte to
;read from, not a raw value.

zero_fill_byte:
  .byte $00